#pragma once

static unsigned int g_VidWidth;

static unsigned int g_VidHeight;

static unsigned short g_BitCount;

static GLuint g_texid;			// Handle to texture

static GLuint g_buffer;			// Handle to buffer

static HWND g_hWnd;				// Handle of the window

static HINSTANCE g_hInstance;	// Instance of the window

static HDC g_hDC;				// Handle of the device context

static HGLRC g_hRC;				// Handle of the OpenGL rendering context

static BOOL g_ExitThreadVar;	// Variable by which help the thread can be asked to exit

static BOOL g_syncVar;			// Variable for synchronizing the threads

static const void *g_pVidData;	// Pointer to the video data

static HANDLE g_WindowProcThreadHandle;	// Handle of the thread receiving window messages

static DWORD g_WindowProcThreadID;		// Id of the thread receiving window messages

LRESULT WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

DWORD Create_Window(LPVOID lpThreadParameter);

HRESULT Create_OpenGLTexBuffers(void);

HRESULT MainCUDAProc();

DWORD Create_Window(LPVOID lpThreadParameter);

DWORD CudaThreadProc(LPVOID lpThreadParameter)
{
	// Initialization
	g_ExitThreadVar	= FALSE;
	g_syncVar		= FALSE;

	
	wglMakeCurrent(g_hDC, g_hRC);

	Create_OpenGLTexBuffers();									// Create OpenGL texture and buffer

	while(true) {
		SuspendThread(*((HANDLE*)lpThreadParameter));
		if(g_ExitThreadVar == TRUE)
			break;
		MainCUDAProc();
		g_syncVar = TRUE;
	};
	// Cleanup
	glDeleteBuffers(1, &g_buffer);
	glDeleteTextures(1, &g_texid);
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(g_hRC);
	ReleaseDC(g_hWnd, g_hDC);
	PostThreadMessage(g_WindowProcThreadID, WM_DESTROY, 0, 0);
	UnregisterClass ("OpenGLWindowClass", g_hInstance);
	DeleteDC(g_hDC);
	g_syncVar = TRUE;
	return 0;
};

DWORD Create_Window(LPVOID lpThreadParameter)
{
	int pf;
	PIXELFORMATDESCRIPTOR pfd;
	// Register the windows class
	LPTSTR T_ClassNampe = TEXT("OpenGLWindowClass");
	WNDCLASS wndClass;
	wndClass.style = CS_DBLCLKS;
	wndClass.lpfnWndProc = WindowProc;		// specifies method to handle event messages
	wndClass.cbClsExtra = 0;
	wndClass.cbWndExtra = 0;
	wndClass.hInstance = g_hInstance;
	wndClass.hIcon = NULL;
	wndClass.hCursor = LoadCursor( NULL, IDC_ARROW );
	wndClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndClass.lpszMenuName = NULL;
	wndClass.lpszClassName = TEXT("OpenGLWindowClass");

	if( !RegisterClass(&wndClass))
	{
		DWORD dwError = GetLastError();
		if( dwError != ERROR_CLASS_ALREADY_EXISTS ) {
			MessageBox(NULL, TEXT("RegisterClass"), TEXT("Failure Code:")/*HRESULT_FROM_WIN32(dwError)*/, MB_OK | MB_ICONERROR);
			return dwError;
		};
	};

	// Create the render window
	g_hWnd = CreateWindow(TEXT("OpenGLWindowClass"), TEXT("Cuda Video Renderer"), WS_OVERLAPPEDWINDOW,
									0, 0, g_VidWidth, g_VidHeight, NULL,
									NULL, g_hInstance, NULL);
	if(g_hWnd == NULL)
	{
		DWORD dwError = GetLastError();
		MessageBox(NULL, TEXT("CreateWindow"), TEXT("Failure Code:")/*HRESULT_FROM_WIN32(dwError))*/, MB_OK | MB_ICONERROR);
		return dwError;
	};

	g_hDC = GetDC(g_hWnd);				// Create device context

	memset(&pfd, 0, sizeof(pfd));
    pfd.nSize        = sizeof(pfd);
    pfd.nVersion     = 1;
    pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType   = PFD_TYPE_RGBA;
    pfd.cColorBits   = 32;

    pf = ChoosePixelFormat(g_hDC, &pfd);
    if (pf == 0) {
	MessageBox(NULL, "ChoosePixelFormat() failed:  "
		   "Cannot find a suitable pixel format.", "Error", MB_OK); 
	return 0;
    } 
 
    if (SetPixelFormat(g_hDC, pf, &pfd) == FALSE) {
	MessageBox(NULL, "SetPixelFormat() failed:  "
		   "Cannot set format specified.", "Error", MB_OK);
	return 0;
    } 

    DescribePixelFormat(g_hDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

    g_hRC = wglCreateContext(g_hDC);	// Create OpenGL rendering context
    wglMakeCurrent(g_hDC, g_hRC);		//
	glClear(GL_COLOR_BUFFER_BIT);

    ShowWindow(g_hWnd, SW_SHOW);
	wglMakeCurrent(NULL, NULL);

	MSG msg;
	BOOL l_bRet;

	g_syncVar = TRUE;

	while(l_bRet = GetMessage(&msg, NULL, 0, 0) != 0)
    {
		if(l_bRet == -1) {
			ASSERT(l_bRet != -1);
			break;
		}
		else if(msg.message == WM_DESTROY) {		// WM_DESTROY is sent by another thread to advise this thread
			DestroyWindow(g_hWnd);					// to destroy the window
		}
		else {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		};
    };
    return (DWORD)msg.wParam;
};


LRESULT WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
};

HRESULT Create_OpenGLTexBuffers(void)
{
	glewInit();
	glEnable(GL_TEXTURE_2D);
	glGenTextures(1, &g_texid);
    glBindTexture(GL_TEXTURE_2D, g_texid);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, g_VidWidth, g_VidHeight, 
                 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    
    glGenBuffers(1, &g_buffer);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, g_buffer);
	glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, 
                    g_VidWidth*g_VidHeight*(g_BitCount/8), 
                    NULL, GL_STREAM_COPY);
	return S_OK;
};

HRESULT MainCUDAProc()
{
	void *l_pTexBuffer_D;
	cudaGLRegisterBufferObject(g_buffer);
	cudaGLMapBufferObject(&l_pTexBuffer_D, g_buffer);
	cudaMemcpy(l_pTexBuffer_D, g_pVidData, g_VidWidth*g_VidHeight*(g_BitCount/8), cudaMemcpyHostToDevice);
	cudaGLUnmapBufferObject(g_buffer);
	cudaGLUnregisterBufferObject(g_buffer);

	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, g_VidWidth, g_VidHeight, 
					GL_RGBA, GL_UNSIGNED_BYTE, 0);
	glBegin(GL_QUADS);
	glVertex2f(-1, -1); glTexCoord2f(0, 0);
	glVertex2f(-1, 1); glTexCoord2f(1, 0);
	glVertex2f(1, 1); glTexCoord2f(1, 1);
	glVertex2f(1, -1); glTexCoord2f(0, 1);
	glEnd();
	SwapBuffers(g_hDC);
	return S_OK;
};