#include "CUDA_VideoRenderer.h"
#include "Registry.h"
#include <GL/glew.h>			// OpenGL extension wrangler
#include <cutil.h>
#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>	// Header for cooperation between CUDA and OpenGL
#include "CUDA.h"

CUDA_VideoRenderer::CUDA_VideoRenderer(LPUNKNOWN pUnk, HRESULT *phr)
:	CBaseVideoRenderer(CLSID_DS_CUDA_RENDERER, NAME("CUDA_VideoRenderer"), pUnk, phr)
{
	
};

CUDA_VideoRenderer::~CUDA_VideoRenderer()
{

};

CUnknown* WINAPI CUDA_VideoRenderer::CreateInstance(LPUNKNOWN pUnkOuter, HRESULT *phr)
{
	CBaseVideoRenderer *pNewObject = new CUDA_VideoRenderer(pUnkOuter, phr);

    if (pNewObject == NULL) {
        *phr = E_OUTOFMEMORY;
    }
    return pNewObject;
};

HRESULT CUDA_VideoRenderer::CheckMediaType(const CMediaType *pmt)
{
	CheckPointer(pmt, E_POINTER);
	if(pmt->majortype != MEDIATYPE_Video)		// If it is not a video type
		return E_INVALIDARG;					// return with an error
	
	if ((pmt->formattype != FORMAT_VideoInfo)||	// If VideoInfoHeader is not available
		(pmt->cbFormat < sizeof(VIDEOINFOHEADER)) ||
		(pmt->pbFormat == NULL))
		return E_INVALIDARG;					// return with an error

	if(pmt->subtype == MEDIASUBTYPE_RGB32)		// Media suptypes being accepted by the filter
		return S_OK;

	return S_FALSE;
};

HRESULT CUDA_VideoRenderer::SetMediaType(const CMediaType *pmt)
{
	CheckPointer(pmt, E_POINTER);
	if ((pmt->formattype != FORMAT_VideoInfo)||	// If VideoInfoHeader is not available
		(pmt->cbFormat < sizeof(VIDEOINFOHEADER)) ||
		(pmt->pbFormat == NULL))
		return E_INVALIDARG;					// return with an error
	const VIDEOINFOHEADER *pviBmp = reinterpret_cast<VIDEOINFOHEADER*>(pmt->Format());  //Bitmap info header
	g_VidWidth		= pviBmp->bmiHeader.biWidth;			// Width of the bitmap, in pixels
	g_VidHeight		= abs(pviBmp->bmiHeader.biHeight);		// Height of the bitmap, in pixels
	g_BitCount		= pviBmp->bmiHeader.biBitCount;			// Number of bits per quantum of a pixel (bpp)

	g_syncVar = FALSE;
	g_WindowProcThreadHandle = CreateThread(NULL, 0, Create_Window, NULL, 0, &g_WindowProcThreadID);		// Creating Window
	while(g_syncVar == FALSE)
		Sleep(1);

	g_syncVar = FALSE;
	m_hCudaThread = CreateThread(NULL, 0, CudaThreadProc, (LPVOID)&m_hCudaThread, CREATE_SUSPENDED, NULL);
	ResumeThread(m_hCudaThread);

	return S_OK;
};

HRESULT CUDA_VideoRenderer::DoRenderSample(IMediaSample *pMediaSample)
{
	CheckPointer(pMediaSample, E_POINTER);
	pMediaSample->GetPointer((BYTE**)&g_pVidData);
	g_syncVar = FALSE;
	ResumeThread(m_hCudaThread);
	while(g_syncVar == FALSE)
		Sleep(1);
	return S_OK;
};

HRESULT CUDA_VideoRenderer::BreakConnect(void)
{
	g_ExitThreadVar = TRUE;
	g_syncVar = FALSE;
	ResumeThread(m_hCudaThread);
	while(g_syncVar == FALSE)
		Sleep(1);
	return CBaseRenderer::BreakConnect();
};
